/*****************************************************************************
File:    dploader.c
Purpose: Reset Hostess i and start Turbo Debugger Remote if needed.
         Load a binary file into the Hostess i dual port RAM and
         signal the COM processor to begin execution.
Company: Comtrol Corporation
Author:  Craig Harrison
Release: 1.00: February 4, 1988 - Craig Harrison - Original
         1.01: June 23, 1991 - Craig Harrison
               1. Add support for HOSTESS i control register initialization.
         1.02: Aug  23, 1991 - Craig Harrison
               1. Add reset option
               2. Add Turbo Debugger Remote startup option
         1.03: 12-3-91 - Craig Harrison
               Changed all "outportb" to "outp" for compatability with
               Microsoft C.
         2.0:  9-15-92 - Craig Harrison
               Changed download to use new firmware utilities and new
               firmware user area location.  Delete option to strip
               first part of download file, this not needed with
               new entry point having offset of 0. Delete questions for
               addressing, these now defined in DPRAM.H.  Correct CR
               numbers, they were 1 less then should be.  Updated cr4[]
               for IRQ disable bit.
         2.0.1: 10-21-92 - Craig Harrison
                Removed 64K download file size limit.
         2.1:  1-28-93 - Craig Harrison
               1. Function download() set interaction flag twice - removed
                  first occurance.
*******************************************************************************
Copyright 1988-1993 Comtrol Corporation. All rights reserved.
Subject to developers license agreement.
*******************************************************************************
*/
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <stdlib.h>
#include <dos.h>
#include "dpram.h"
#include "firmuser.h"

/* Control register defines */
#define EN_RAM    0x00        /* Enable DPRAM */
#define CR1       0x01        /* Select Control Reg 1 */
#define CR2       0x02        /* Select Control Reg 2 */
#define DIS_RAM   0x04        /* Disable DPRAM */
#define CR3       0x08        /* Select Control Reg 3 */
#define CR4       0x10        /* Select Control Reg 4 */

/* CR2 values for below 1 Meg addresses */
#define ADD80000 0x20
#define ADD90000 0x24
#define ADDA0000 0x28
#define ADDB0000 0x2c
#define ADDC0000 0x30
#define ADDD0000 0x34
#define ADDE0000 0x38
#define ADDF0000 0x3c

/* Firmware utility defines */
#define COPY      0x02        /* firmware utility copy command */
#define EXEC      0x01        /* firmware utility execute ctrlpgm command */
#define DONE      0x01        /* firmware utility DONE status flag */



FIRMUSER_T far *fu_p;         /* ptr to firmuser user area */
int io;                       /* controller base I/O address */

/* CR3 values for various window sizes and offsets */
char w16k[8] = {0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f};  /* 16K windows */
char w32k[4] = {0x30,0x32,0x34,0x36};            /* 32K windows */
char w64k[2] = {0x20,0x24};                      /* 64K windows */

/* CR4 values for IRQ 0-15, 0 means IRQ is disabled */
char cr4[16] = {0,0,0,0x08,0x09,0x0a,0,0,0,0x0b,0x0c,0x0d,0x0e,0,0,0x0f};

void cr_init(void);
void download(void);
void reset_board(void);
int invoke_tdrem(void);


main()
{
   fu_p = (FIRMUSER_T far *)FIRMUSER_ADDR_SYS; /* firmuser area address */
   io = IO_SYS;                        /* I/O base address */
   reset_board();                      /* reset board if needed */
   download();                         /* download & start ctrlpgm */
   return(0);
}

/*****************************************************************************
Purpose: Reset board, initialize control registers.
Entry: Nothing
Returns: Nothing
*/
void reset_board(void)
{
   char strbuff[81];
   int i;
   unsigned rec_flag;

   while(1)
   {
      printf("Reset HOSTESS i controller (Y/N)? ");
      scanf("%s",strbuff);
      switch(*strbuff)
      {
         case 'y':
         case 'Y':
            outp(io+3,0);
            delay(100);
            outp(io+3,0xff);
            printf("   Waiting for reset to complete\n");
            delay(500);
            cr_init();
            for(i = 0;i < 21;i++)
            {
               delay(1000);
               if(fu_p->i_flag == 0x55aa)
                  break;
            }
            if(((rec_flag = fu_p->i_flag) != 0x55aa) && i >= 21)
            {
               printf("   Reset failed, interaction flag = %XH\n",rec_flag);
               exit(1);
            }
            return;
         case 'n':
         case 'N':
            cr_init();
            return;
         default:
            break;
      }
   }
}

/*****************************************************************************
Purpose: Load a binary file into dual port memory using the firmware
         utilities, signal the COM Processor to start execution,
         verify execution started.
Entry:   Nothing
Returns: Nothing
Notes:   The file to be downloaded must have an entry point of c0:0 and
         must write 55aaH to the interaction flag.
         The download uses the firmware utilities, which require a source
         buffer in DPRAM (dual port RAM).  This source buffer is 256 bytes
         in size, and is located at the control program's entry point.
         The utility reuses the same source buffer for each 256 byte block
         of data downloaded, so when finished the first 256 bytes of the
         control program must be restored.  This is  done by saving the
         first 256 bytes in sbuf[], then writing it into DPRAM as the last
         step of the download.
*/
void download(void)
{
   char far *ram_p;
   char fname[13];                     /* Name of file to download */
   unsigned wait;                      /* Do nothing loop counter */
   unsigned rec_flag;                  /* Received copy of interact flag*/
   unsigned long absaddr;              /* Download absolute address */
   static char buf[256];               /* Buffer for data to be xfered
                                          to dual port RAM */
   static char sbuf[256];              /* save buffer for 1st 256 bytes of
                                          of data to xfer to DPRAM */
   int sbytes;                         /* number bytes saved in sbuf[] */
   long int nbytes;                    /* Number of bytes downloaded */
   int retry,                          /* Loop control flag */
       i,
       f_h,                            /* File handle */
       bytesread;                      /* Bytes read on current read */

   /* Check interaction flag */
   if((rec_flag = fu_p->i_flag) != 0x55aa)
   {
      printf("Bad interaction flag = %XH before download\n",rec_flag);
      exit(1);
   }

   /* Get file name and open file for read */
   retry = 1;
   while(retry)
   {
      printf("Enter control program file name to download: ");
      scanf("%s12",fname);
      if((f_h = open(fname,O_BINARY)) == -1)
         printf("   Unable to open file %s, try again.\n",fname);
      else
         retry = 0;
   }

   /* Read first 256 bytes of ctrlpgm into save buffer because the first 256
      bytes in DPRAM will be used as the source buffer */
   printf("Downloading %s ... \n",fname);
   bytesread = read(f_h,sbuf,256);
   if(bytesread <= -1)                 /* Error reading file */
   {
      printf("   Error reading file %s\n",fname);
      exit(1);
   }
   nbytes = bytesread;
   sbytes = bytesread;

   /* Set up source buf address in firmware user area for copy command */
   absaddr = (unsigned long)(ENTRY_SEG << 4) + ENTRY_OFF;
   fu_p->cmd = COPY;            /* firmware utility copy command */
   fu_p->msg[0] = (unsigned)ENTRY_SEG;   /* source buffer segment */
   fu_p->msg[1] = (unsigned)ENTRY_SEG >> 8;
   fu_p->msg[2] = (unsigned)ENTRY_OFF;   /* source buffer offset */
   fu_p->msg[3] = (unsigned)ENTRY_OFF >> 8;

   /* Download all but first 256 byte block */
   while(!eof(f_h))
   {
      ram_p = (char far *)LOAD_ADDR_SYS; /* Source buf at ctrlpgm entry pnt */
      absaddr += 256;                  /* Next destination address */
      bytesread = read(f_h,buf,256);   /* Read 256 bytes into buffer */
      if(bytesread <= -1)              /* Error reading file */
      {
         printf("   Error reading file %s\n",fname);
         exit(1);
      }
      nbytes += bytesread;
      for(i = 0;i < bytesread;i++) /* Move buffer into DPRAM source buffer */
      {
         *ram_p++ = buf[i];
      }

      /* Firmware utility copy of DPRAM source buf bytes to destination */
      fu_p->status = 0;             /* Clear status */
      fu_p->msg[4] = absaddr >> 4;  /* Destination segment */
      fu_p->msg[5] = absaddr >> 12;
      fu_p->msg[6] = absaddr & 0x0000000fL;  /* Destination offset */
      fu_p->msg[7] = 0;
      fu_p->msg[8] = bytesread;
      fu_p->msg[9] = bytesread >> 8;
      outp(io + 2,0);                  /* Interrupt COM Processor */
      for(wait = 0;wait < 10;wait++)   /* Wait for copy to finish */
      {
         if(fu_p->status == DONE)
            break;
         delay(2);
      }
      if((wait >= 10) && (fu_p->status != DONE)) /* Copy failed */
      {
         printf("Firmware copy command failed\n");
         exit(1);
      }
   }

   /* Download first 256 bytes of control program to DPRAM, this area was used
      as the source buffer for the firmware utility copies */
   ram_p = (char far *)LOAD_ADDR_SYS;
   for(i = 0;i < sbytes;i++)
   {
      *ram_p++ = sbuf[i];
   }

   /* Download successful, start control program by interrupting COM Processor
      using firmware utility */
   if(close(f_h) != 0)           /* close ctrlpgm file */
   {
      printf("   Error closing file %s after reading into DPRAM\n",fname);
      exit(1);
   }
   printf("%ld bytes downloaded successfully\n",nbytes);
   if(invoke_tdrem())                  /* Turbo Debugger invoked */
      return;                          /* don't start ctrlpgm */
   fu_p->i_flag = 0xaa55;              /* set up interaction flag */
   fu_p->cmd = EXEC;             /* firmware utility command to exec ctrlpgm */
   fu_p->status = 0;
   fu_p->msg[0] = (unsigned)ENTRY_SEG;   /* entry point segment */
   fu_p->msg[1] = (unsigned)ENTRY_SEG >> 8;
   fu_p->msg[2] = (unsigned)ENTRY_OFF;   /* entry point offset */
   fu_p->msg[3] = (unsigned)ENTRY_OFF >> 8;
   outp(io + 2,0);                     /* Interrupt COM Processor */
   printf("COM processor interrupted to start control program\n");
   for(wait = 0;wait < 10;wait++)      /* wait for control program to */
   {                                   /*   restore interaction flag */
      delay(400);
      if(fu_p->i_flag == 0x55aa)
         break;
   }
   if((wait >= 10) && ((rec_flag = fu_p->i_flag) != 0x55aa))
   {
      printf("Control program failed to start, interact flag = %XH\n",
              rec_flag);
      exit(1);
   }
   else
   {
      printf("Control program started execution\n");
   }
}

/*****************************************************************************
Purpose: HOSTESS i control register initialization
Entry: Nothing
Returns: Nothing
*/
void cr_init(void)
{
   outp(io+1,CR1 | DIS_RAM);                /* access CR1, PC mode */
   outp(io,0);                              /* below 1 Meg */
   outp(io+1,CR2 | DIS_RAM);                /* access CR2 */
   outp(io,ADDD0000);                       /* below 1 Meg RAM address */
   outp(io+1,CR3 | DIS_RAM);                /* access CR3 */
   outp(io,w64k[0]);                        /* 64K lower window */
   outp(io+1,CR4 | DIS_RAM);                /* access CR4 */
   outp(io,cr4[0]);                         /* no IRQ set */
   outp(io+1,EN_RAM);                       /* enable DPRAM */
}

/*****************************************************************************
Purpose: Invoke Turbo Debugger Remote kernel, if needed.
Entry: Nothing
Returns: 1 if tdrem is invoked, else 0
*/
int invoke_tdrem()
{
   char strbuff[0x81];                  /* Stores string entered by user */
   int do_tdrem,valid_answer,wait;
   unsigned save_word;                  /* Saved copy of 1st ctrlpgm word */
   unsigned far *inst_p;               /* Ptr to control pgm entry point */

   valid_answer = 0;
   while(!valid_answer)
   {
      printf("Invoke Turbo Debugger Remote Kernel on HOSTESS i controller (Y/N)? ");
      scanf("%s",strbuff);
      switch(*strbuff)
      {
         case 'y':
         case 'Y':
            valid_answer = 1;
            do_tdrem = 1;
            break;
         case 'n':
         case 'N':
            valid_answer = 1;
            do_tdrem = 0;
            break;
         default:
            break;
      }
   }
   if(do_tdrem)
   {
      inst_p = (unsigned far *)LOAD_ADDR_SYS; /* Ptr to ctrl pgm entry point */
      save_word = *inst_p;             /* Save 1st word of ctrlpgm */
      *inst_p = 0x27cd;                  /* Write int 27H instruction used to
                                          invoke TDREM kernel */
      fu_p->cmd = EXEC;           /* firmware utility command to exec ctrlpgm */
      fu_p->status = 0;
      fu_p->msg[0] = (unsigned)ENTRY_SEG;   /* entry point segment */
      fu_p->msg[1] = (unsigned)ENTRY_SEG >> 8;
      fu_p->msg[2] = (unsigned)ENTRY_OFF;   /* entry point offset */
      fu_p->msg[3] = (unsigned)ENTRY_OFF >> 8;
      outp(io + 2,0);                  /* Interrupt COM Processor */
      delay(10);                        /* Give it a chance to execute */
      printf("Turbo Debugger Remote Kernel Started\n");
      *inst_p = save_word;             /* Restore 1st word of ctrlpgm */
   }
   return(do_tdrem);
}

